/** @file tokenwriter.h @brief Class TokenWriter */

/* FAU Discrete Event Systems Library (libfaudes)

Copyright (C) 2006  Bernd Opitz
Exclusive copyright is granted to Klaus Schmidt

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#ifndef TOKENWRITER_H

#include "definitions.h"
#include "token.h"
#include <fstream>
#include <iostream>

namespace faudes {

/**
 * A TokenWriter writes sequential tokens to a file, a string or stdout. It is the counterpart
 * of the TokenReader. Since wrtiting data comparatively straight foreward, there is no explicit 
 * support of sections etc. It is left to the calling function to organise the output. 
 */

class TokenWriter {

 public:

  /**
   * Mode of operation: write to from file, string or stdout
   */
  enum Mode {File, Stdout, String};

  /**
   * Console or String TokenWriter constructor
   *
   * @exception Exception
   *   - faudes::Exception ios error opening file (id 2)
   */
  TokenWriter(Mode mode);

  /** 
   * File TokenWriter constructor
   *
   * @param rFilename
   *   File to write
   * @param openmode
   *   std::ios::openmode
   * @exception Exception
   *   - faudes::Exception ios error opening file (id 2)
   */
  TokenWriter(const std::string& rFilename, 
	      std::ios::openmode openmode = std::ios::out|std::ios::trunc);

  /** Destructor. Calls close */
  ~TokenWriter(void);

  /** 
   * Retrieve output as string (if in String mode)
   *
   * @exception Exception
   *   - faudes::Exception not in string mode (id 2)
   */
  std::string Str(void);
	
  /**
   * Get number of columns in a line
   *
   * @return
   *   # of columns in a line
   */
  int Columns(void) const;

  /**
   * Set number of columns in a line 
   *
   * @param columns
   *   # of columns in a line
   */
  void Columns(int columns);

  /**
   * Write endl separator (obmitted in String mode)
   *
   * @exception Exception
   *   - faudes::Exception ios error writing file (id 2)
   */
  void Endl(void);

  /**
   * Write next token
   *
   * @param rToken
   *   Token to write
   * @exception Exception
   *   - faudes::Exception ios error wrtiting file (id 2)
   */
  void Write(Token& rToken);

  /**
   * Write name 
   *
   * @param rString
   *   String to write
   * @exception Exception
   *   - faudes::Exception ios error wrtiting file (id 2)
   */
  void Write(const std::string& rString);

  /**
   * Write index
   *
   * @param index
   *   Index to write
   * @exception Exception
   *   - faudes::Exception ios error wrtiting file (id 2)
   */
  void Write(Idx index);

  /**
   * Write float
   *
   * @param val
   *   float to write
   * @exception Exception
   *   - faudes::Exception ios error wrtiting file (id 2)
   */
  void WriteFloat(const double& val);

  /**
   * Write option
   *
   * @param rOpt
   *   option to write
   * @exception Exception
   *   - faudes::Exception ios error wrtiting file (id 2)
   */
  void WriteOption(const std::string& rOpt);

  /**
   * Write begin label
   * 
   * @param rLabel
   *   End label, e.g. "Alphabet"
   * @exception Exception
   *   - faudes::Exception ios error wrtiting file (id 2)
   */
  void WriteBegin(const std::string& rLabel);

  /**
   * Write end label
   * 
   * @param rLabel
   *   End label, e.g. "Alphabet"
   * @exception Exception
   *   - faudes::Exception ios error wrtiting file (id 2)
   */
  void WriteEnd(const std::string& rLabel);
	
  /**
   * Write comment
   * 
   * @param rComment
   *   Comment to write
   * @exception Exception
   *   - faudes::Exception ios error wrtiting file (id 2)
   */
  void Comment(const std::string& rComment);

  /** 
   * Operator for writing tokens
   *
   * @param rToken
   *   Token to write
   * @exception Exception
   *   - faudes::Exception ios error wrtiting file (id 2)
   */
  void operator << (Token& rToken) {
    Write(rToken);
  }

  /** 
   * Operator for writing std::strings to a stream 
   *
   * @param rString
   *   String to write
   * @exception Exception
   *   - faudes::Exception ios error wrtiting file (id 2)
   */
  void operator << (const std::string& rString) {
    Write(rString);
  }

  /** 
   * Operator for writing Idxs to a stream
   *
   * @param index
   *   Index to write
   * @exception Exception
   *   - faudes::Exception ios error wrtiting file (id 2)
   */
  void operator << (const Idx index) {
    Write(index);
  }


 private:
  /** output mode */
  Mode mMode;

  /** ostream object pointer*/
  std::ostream* mpStream;

  /** actual stream object, file output */
  std::ofstream mFStream;

  /** actual stream object, string output */
  std::ostringstream mSStream;

  /** Filename */
  std::string mFileName;

  /** Number of columns */
  int mColumns;

  /** Column counter */
  int mColCount;

};

} // namespace faudes

#define TOKENWRITER_H
#endif 

