/** @file cgenerator.h Classes TcGenerator, cGenerator and AttributeCFlags */

/* FAU Discrete Event Systems Library (libfaudes)

Copyright (C) 2006  Bernd Opitz
Copyright (C) 2007  Thomas Moor
Exclusive copyright is granted to Klaus Schmidt

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#ifndef CGENERATOR_H

#include "definitions.h"
#include "agenerator.h"

namespace faudes {


/**
 * Attribute class to model event controllability properties.
 *
 * This attribute is meant to be an event attribute and can distinguish between
 * controllable, observable and forcible events. It is based on faudes::AttributeFlags
 * and uses the lower three bits in the flag word to store the respective boolean values.
 * The AttributeCFlags class adds convenience functions to access these bits and a default-value 
 * that corresponds to observable and neiter controllable nor forcible.
 *
 * Presuming that only controllability flags are uses (different from default), the
 * token representation is by an Option String consisting of the initials c,o and f,
 * where initials are capitatised for set flags and default values are not writtem; eg +C+ 
 * for a controllable event tzhat is observable (default) and not forcible (default).
 * If other than the three controllability bits are used, std. hex format is used.
 * 
 */


class AttributeCFlags : public AttributeFlags {
 public:
  /**
   * Default constructor 
   */
  AttributeCFlags(void) {mFlags=mDefCFlags;}

  /** Destructor */
  virtual ~AttributeCFlags(void) {};

  /**
   * Read attribute value from tokenreader.
   *
   * Reads a token if it can be interpreted as AttributeCFlag.
   * No token mismatch exceptions are thrown on error.
   *
   * @param rTr
   *   TokenReader to read from
   *
   * @exception Exception
   *   - IO error (id 1)
   */
  virtual void Read(TokenReader& rTr);
 
  /**
   * Write attribute value tokenwriter
   *
   * @param rTw
   *   TokenWriter to write to
   *
   * @exception Exception
   *   - IO error (id 2)
   */
  virtual void Write(TokenWriter& rTw) const;

  /**
   * Write attribute value to console
   *
   */
  virtual void Write(void) const;

  /**
   * Write attribute value to string
   *
   * @return
   *  String representation of attribute value
   *
   * @exception Exception
   *   - IO error (id 2)
   */
  virtual std::string ToString(void) const;

  /**
   * Set controllable flag 
   */

  void SetControllable(void) { mFlags |= mControllableFlag; }

  /**
   * Clear controllable flag 
   */

  void ClrControllable(void) { mFlags &= ~mControllableFlag; };
     
  /**
   * Query controllablility 
   */

  bool Controllable(void) const {return ( (mFlags & mControllableFlag) != 0 ); }


  /**
   * Set observable flag 
   */

  void SetObservable(void) { mFlags |= mObservableFlag; }

  /**
   * Clear observable flag 
   */

  void ClrObservable(void) { mFlags &= ~mObservableFlag; };
     
  /**
   * Query observablility 
   */

  bool Observable(void) const {return ( (mFlags & mObservableFlag) != 0 ); }


  /**
   * Set forcible flag 
   */

  void SetForcible(void) { mFlags |= mForcibleFlag; }

  /**
   * Clear forcible flag 
   */

  void ClrForcible(void) { mFlags &= ~mForcibleFlag; };
     
  /**
   * Query forcibility 
   */

  bool Forcible(void) const {return ( (mFlags & mForcibleFlag) != 0 ); }


  /** 
   * Test for default value
   */
  bool  IsDefault(void) const {return mFlags==mDefCFlags;};

  // flag masks for the three properties
  const static fType mControllableFlag=0x01;
  const static fType mObservableFlag  =0x02;
  const static fType mForcibleFlag    =0x04;

 private:
  /** Overall default value */
  const static fType mDefCFlags         =0x02;

  /** All flags used by CFlags */
  const static fType mAllCFlags         =0x07;

}; // class AttributeCFlags


/**
 * Generator with controllability attributes. The cGenerator is a variant of the 
 * Generator to add an interface for events with controllabilty attributes, ie
 * an event may be controllable, observable or forcible. 
 *
 * Technically, the construct is based on the specialized attribute class faudes::AttributeCFlags
 * that provides attributes with semantics for controllability properties. The TcGenerator
 * expects an event attribute template parameter with the minimum interface defined in AttribueCFlags.
 * Thus, you can add further semantics by deriving a class AttributeCFlagsAndMore  from
 * AttribueeCFlags and use this as event attribute parameter for TcGenerator. To model
 * a plain finite state machine plus controllability properties, use TcGenerator with
 * AttributeCFlags for the event attribute parameter and AttributeVoid for the other
 * parameters. For convenience, this has been typedefed as cGenerator.
 *
 * @ingroup GeneratorClasses
 */

template <class GlobalAttr, class StateAttr, class EventAttr, class TransAttr>
    class TcGenerator : public TaGenerator<GlobalAttr, StateAttr, EventAttr, TransAttr> {    
  public:
    /**
     * Creates an emtpy cGenerator object 
     */
    TcGenerator(void);

    /** 
     * cGenerator from a std Generator. Copy constructor 
     *
     * @param rOtherGen
     */
    TcGenerator(const vGenerator& rOtherGen);
        
    /** 
     * cGenerator from a cGenerator. Copy constructor 
     *
     * @param rOtherGen
     */
    TcGenerator(const TcGenerator& rOtherGen);

    /**
     * construct a cGenerator from file
     *
     * @param pFileName
     *   Filename
     *
     * @exception Exception
     *   If opening/reading fails an Exception object is thrown (id 1, 50, 51)
     */
    TcGenerator(const char* pFileName);

    /**
     * construct on heap
     *
     * @return 
     *   new Generator 
     */
     TcGenerator* NewP(void) const;

    /**
     * construct on stack
     *
     * @return 
     *   new Generator 
     */
     TcGenerator NewCGen(void) const;

    /**
     * Add an existing controllable event to generator. 
     * An entry in the global eventtable will be made.
     *
     * @param index
     *   Event index
     */
    void InsControllableEvent(Idx index);

    /**
     * Add new named controllable event to generator. 
     * An entry in the global eventtable will be made if event is new.
     *
     * @param rName
     *   Name of the event to add
     *
     * @return 
     *   New global unique index
     */
    Idx InsControllableEvent(const std::string& rName);
    
    /**
     * Add an existing uncontrollable event to generator. 
     * An entry in the global eventtable will be made.
     *
     * @param index
     *   Event index
     */
    void InsUncontrollableEvent(Idx index);

    /**
     * Add new named uncontrollable event to generator. 
     * An entry in the global eventtable will be made if event is new.
     *
     * @param rName
     *   Name of the event to add
     *
     * @return 
     *   New global unique index
     */
    Idx InsUncontrollableEvent(const std::string& rName);
        
    /**
     * Mark event controllable (by index)
     *
     * @param index
     *   Event index
     */
    void SetControllable(Idx index);

    /** 
     * Mark event controllable (by name)
     *
     * @param rName
     *   Event name    
     */
    void SetControllable(const std::string& rName);

    /** 
     * Mark set of events controllable (by index)
     *
     * @param rEvents
     *   EventSet
     */
    void SetControllable(const EventSet& rEvents);
        
    /**
     * Mark event uncontrollable (by index)
     * 
     * @param index
     *   Event index
     */
    void ClrControllable(Idx index);

    /**
     * Mark event uncontrollable (by name)
     * 
     * @param rName
     *   Event name
     */
    void ClrControllable(const std::string& rName);

    /** 
     * Mark set of events uncontrollable (by index)
     *
     * @param rEvents
     *   EventSet
     */
    void ClrControllable(const EventSet& rEvents);
          
    /**
     * Is event controllable (by index)
     *
     * @param index
     *   Event index
     *
     * @return
     *   True / false
     */
    bool Controllable(Idx index) const;

    /**
     * Is event controllable (by name)
     *
     * @param rName
     *   Event name
     *
     * @return
     *   True / false
     */
    bool Controllable(const std::string& rName) const;

    /** 
     * Get EventSet with controllable events
     *
     * @return
     *   EventSet of controllable events
     */
    EventSet ControllableEvents(void) const;

    /**
     * Get EventSet with uncontrollable events
     *
     * @return 
     *   EventSet of uncontrollable events
     */
    EventSet UncontrollableEvents(void) const;

    /**
     * Add an existing observable event to generator. 
     * An entry in the global eventtable will be made.
     *
     * @param index
     *   Event index
     */
    void InsObservableEvent(Idx index);

    /**
     * Add new named observable event to generator. 
     * An entry in the global eventtable will be made if event is new.
     *
     * @param rName
     *   Name of the event to add
     *
     * @return 
     *   New global unique index
     */
    Idx InsObservableEvent(const std::string& rName);
    
    /**
     * Add an existing unobservable event to generator. 
     * An entry in the global eventtable will be made.
     *
     * @param index
     *   Event index
     */
    void InsUnobservableEvent(Idx index);

    /**
     * Add new named unobservable event to generator. 
     * An entry in the global eventtable will be made if event is new.
     *
     * @param rName
     *   Name of the event to add
     *
     * @return 
     *   New global unique index
     */
    Idx InsUnobservableEvent(const std::string& rName);
        
    /**
     * Mark event observable (by index)
     *
     * @param index
     *   Event index
     */
    void SetObservable(Idx index);

    /** 
     * Mark event observable (by name)
     *
     * @param rName
     *   Event name    
     */
    void SetObservable(const std::string& rName);

    /**
     * Mark set of events observable 
     *
     * @param rEvents
     *   EventSet
     */
    void SetObservable(const EventSet& rEvents);
        
    /**
     * Mark event unobservable (by index)
     * 
     * @param index
     *   Event index
     */
    void ClrObservable(Idx index);

    /**
     * Mark event unobservable (by name)
     * 
     * @param rName
     *   Event name
     */
    void ClrObservable(const std::string& rName);

    /**
     * Mark set of events unobservable 
     *
     * @param rEvents
     *   EventSet
     */
    void ClrObservable(const EventSet& rEvents);
          
    /**
     * Is event observable (by index)
     *
     * @param index
     *   Event index
     *
     * @return
     *   True / false
     */
    bool Observable(Idx index) const;

    /**
     * Is event observable (by name)
     *
     * @param rName
     *   Event name
     *
     * @return
     *   True / false
     */
    bool Observable(const std::string& rName) const;

    /** 
     * Get EventSet with observable events
     *
     * @return
     *   EventSet of controllable events
     */
    EventSet ObservableEvents(void) const;

    /**
     * Get EventSet with unobservable events
     *
     * @return 
     *   EventSet of uncontrollable events
     */
    EventSet UnobservableEvents(void) const;

    /**
     * Add an existing forcible event to generator. 
     * An entry in the global eventtable will be made.
     *
     * @param index
     *   Event index
     */
    void InsForcibleEvent(Idx index);

    /**
     * Add new named forcible event to generator. 
     * An entry in the global eventtable will be made if event is new.
     *
     * @param rName
     *   Name of the event to add
     *
     * @return 
     *   New global unique index
     */
    Idx InsForcibleEvent(const std::string& rName);
    
    /**
     * Add an existing unforcible event to generator. 
     * An entry in the global eventtable will be made.
     *
     * @param index
     *   Event index
     */
    void InsUnforcibleEvent(Idx index);

    /**
     * Add new named unforcible event to generator. 
     * An entry in the global eventtable will be made if event is new.
     *
     * @param rName
     *   Name of the event to add
     *
     * @return 
     *   New global unique index
     */
    Idx InsUnforcibleEvent(const std::string& rName);
        
    /**
     * Mark event forcible (by index)
     *
     * @param index
     *   Event index
     */
    void SetForcible(Idx index);

    /** 
     * Mark event forcible (by name)
     *
     * @param rName
     *   Event name    
     */
    void SetForcible(const std::string& rName);

    /**
     * Mark set of events forcible
     *
     * @param rEvents
     *   EventSet
     */
    void SetForcible(const EventSet& rEvents);
        
    /**
     * Mark event unforcible (by index)
     * 
     * @param index
     *   Event index
     */
    void ClrForcible(Idx index);

    /**
     * Mark event unforcible (by name)
     * 
     * @param rName
     *   Event name
     */
    void ClrForcible(const std::string& rName);

    /**
     * Mark set of events unforcible 
     *
     * @param rEvents
     *   EventSet
     */
    void ClrForcible(const EventSet& rEvents);
          
    /**
     * Is event forcible (by index)
     *
     * @param index
     *   Event index
     *
     * @return
     *   True / false
     */
    bool Forcible(Idx index) const;

    /**
     * Is event forcible (by name)
     *
     * @param rName
     *   Event name
     *
     * @return
     *   True / false
     */
    bool Forcible(const std::string& rName) const;

    /** 
     * Get EventSet with forcible events
     *
     * @return
     *   EventSet of controllable events
     */
    EventSet ForcibleEvents(void) const;

    /**
     * Get EventSet with unforcible events
     *
     * @return 
     *   EventSet of uncontrollable events
     */
    EventSet UnforcibleEvents(void) const;

  private:

  protected:
}; // end class TcGeneraator

    
/** Convenience typedef for std cGenerator */
typedef TcGenerator<AttributeVoid, AttributeVoid, AttributeCFlags,AttributeVoid> cGenerator;

/** EventSet with controllablity properties */
typedef TaNameSet<AttributeCFlags> cEventSet;

/* convenience access to relevant scopes */
#define THIS TcGenerator<GlobalAttr, StateAttr, EventAttr, TransAttr>
#define BASE TaGenerator<GlobalAttr, StateAttr, EventAttr, TransAttr>
#define TEMP template <class GlobalAttr, class StateAttr, class EventAttr, class TransAttr>


// TcGenerator(void)
TEMP THIS::TcGenerator(void) : BASE() {
  FD_DG("TcGenerator(" << this << ")::TcGenerator()");
}

// TcGenerator(rOtherGen)
TEMP THIS::TcGenerator(const TcGenerator& rOtherGen) : BASE(rOtherGen) {
  FD_DG("TcGenerator(" << this << ")::TcGenerator(rOtherGen)");
}

// TcGenerator(rOtherGen)
TEMP THIS::TcGenerator(const vGenerator& rOtherGen) : BASE(rOtherGen) {
  FD_DG("TcGenerator(" << this << ")::TcGenerator(rOtherGen)");
}

// TcGenerator(pFilename)
TEMP THIS::TcGenerator(const char* pFileName) : BASE(pFileName) {
  FD_DG("TcGenerator(" << this << ")::TcGenerator(pFilename) : done");
}

/*
// NewcGenerator()
TEMP THIS THIS::NewcGenerator(void) const {
  THIS res;
  res.EventSymbolTablep(BASE::mpEventSymbolTable);
  return res;
}
*/

// NewP
TEMP THIS* THIS::NewP(void) const {
  // allocate
  THIS* res = new THIS;
  // fix base data
  res->EventSymbolTablep(BASE::mpEventSymbolTable);
  res->mStateNamesEnabled=BASE::mStateNamesEnabled;
  return res;
}

// NewCGen
TEMP THIS THIS::NewCGen(void) const {
  // call base (fixes by assignment constructor)
  THIS res= BASE::NewAGen();
  return res;
}


  // Controllable(index)
  TEMP bool THIS::Controllable(Idx index) const {
    EventAttr attr=BASE::EventAttribute(index);
    return attr.Controllable();
  } 
	
  // Controllable(rName)
  TEMP bool THIS::Controllable(const std::string& rName) const {
    EventAttr attr=BASE::EventAttribute(rName);
    return attr.Controllable();
  } 
	
  // InsControllableEvent(index)
  TEMP void THIS::InsControllableEvent(Idx index) {
    FD_DG("TcGenerator(" << this << ")::InsControllableEvent(" << index << ")");
    EventAttr attr;
    attr.SetControllable();
    BASE::InsEvent(index,attr);
  } 

  // InsControllableEvent(rName)
  TEMP Idx THIS::InsControllableEvent(const std::string& rName) {
    FD_DG("TcGenerator(" << this << ")::InsControllableEvent(" << rName << ")");
    EventAttr attr;
    attr.SetControllable();
    return BASE::InsEvent(rName,attr);
  } 

  // InsUncontrollableEvent(index)
  TEMP void THIS::InsUncontrollableEvent(Idx index) {
    FD_DG("TcGenerator(" << this << ")::InsUncontrollableEvent(" << index << ")");
    EventAttr attr;
    attr.ClrControllable();
    BASE::InsEvent(index,attr);
  } 

  // InsUncontrollableEvent(rName)
  TEMP Idx THIS::InsUncontrollableEvent(const std::string& rName) {
    FD_DG("TcGenerator(" << this << ")::InsUncontrollableEvent(" << rName << ")");
    EventAttr attr;
    attr.ClrControllable();
    return BASE::InsEvent(rName,attr);
  } 
    
  // SetControllable(index)
  TEMP void THIS::SetControllable(Idx index) {
    FD_DG("TcGenerator(" << this << ")::SetControllable(" << index << ")");
    EventAttr attr=BASE::EventAttribute(index);
    attr.SetControllable();
    BASE::mAlphabet.Attribute(index,attr);
  } 

  // SetControllable(rName)
  TEMP void THIS::SetControllable(const std::string& rName) {
    FD_DG("TcGenerator(" << this << ")::SetControllable(" << rName << ")");
    Idx index;
    try{
      index = BASE::EventIndex(rName);
    }
    catch (Exception) {
      std::stringstream errstr;
      errstr << "event \"" << rName << "\" not found in generator \"" << BASE::Name() << "\"";
      throw Exception("TcGenerator::SetControllable", errstr.str(), 202);
    }
    SetControllable(index);
  }

  //SetControllable(rEvents)
  TEMP void THIS::SetControllable(const EventSet& rEvents) {
    FD_DG("TcGenerator(" << this << ")::SetControllable(rEvents)");
    EventSet::Iterator it;
    for(it=rEvents.Begin(); it!=rEvents.End(); it++) {
      SetControllable(*it);
    }
  }
    
  // ClrControllable(index)
  TEMP void THIS::ClrControllable(Idx index) {
    FD_DG("TcGenerator(" << this << ")::ClrControllable(" << index << ")");
    EventAttr attr=BASE::EventAttribute(index);
    attr.ClrControllable();
    BASE::mAlphabet.Attribute(index,attr);
  } 

  // ClrControllable(rName)
  TEMP void THIS::ClrControllable(const std::string& rName) {
    FD_DG("TcGenerator(" << this << ")::ClrControllable(" << rName << ")");
    Idx index;
    try{
      index = BASE::EventIndex(rName);
    }
    catch (Exception) {
      std::stringstream errstr;
      errstr << "event \"" << rName << "\" not found in generator \"" << BASE::Name() << "\"";
      throw Exception("TcGenerator::ClrControllable", errstr.str(), 202);
    }
    ClrControllable(index);
  }

  //ClrControllable(rEvents)
  TEMP void THIS::ClrControllable(const EventSet& rEvents) {
    FD_DG("TcGenerator(" << this << ")::ClrControllable(rEvents)");
    EventSet::Iterator it;
    for(it=rEvents.Begin(); it!=rEvents.End(); it++) {
      ClrControllable(*it);
    }
  }

  //ControllableEvents()
  TEMP EventSet THIS::ControllableEvents(void) const {
    FD_DG("TcGenerator(" << this << ")::ControllableEvents()");
    EventSet res;
    EventSet::Iterator it;
    for(it=BASE::AlphabetBegin(); it!=BASE::AlphabetEnd(); it++) {
      if(Controllable(*it)) res.Insert(*it);
    }
    return res;
  }
  
  //UncontrollableEvents()
  TEMP
    EventSet THIS::UncontrollableEvents(void) const {
    FD_DG("TcGenerator(" << this << ")::UncontrollableEvents()");
    EventSet res;
    EventSet::Iterator it;
    for(it=BASE::AlphabetBegin(); it!=BASE::AlphabetEnd(); it++) {
      if(!Controllable(*it)) res.Insert(*it);
    }
    return res;
  }
  
  // Observable(index)
  TEMP bool THIS::Observable(Idx index) const {
    EventAttr attr=BASE::EventAttribute(index);
    return attr.Observable();
  } 
	
  // Observable(rName)
  TEMP bool THIS::Observable(const std::string& rName) const {
    EventAttr attr=BASE::EventAttribute(rName);
    return attr.Observable();
  } 
	
  // InsObservableEvent(index)
  TEMP void THIS::InsObservableEvent(Idx index) {
    FD_DG("TcGenerator(" << this << ")::InsObservableEvent(" << index << ")");
    EventAttr attr;
    attr.SetObservable();
    BASE::InsEvent(index,attr);
  } 

  // InsObservableEvent(rName)
  TEMP Idx THIS::InsObservableEvent(const std::string& rName) {
    FD_DG("TcGenerator(" << this << ")::InsObservableEvent(" << rName << ")");
    EventAttr attr;
    attr.SetObservable();
    return BASE::InsEvent(rName,attr);
  } 

  // InsUnobservableEvent(index)
  TEMP void THIS::InsUnobservableEvent(Idx index) {
    FD_DG("TcGenerator(" << this << ")::InsUnobservableEvent(" << index << ")");
    EventAttr attr;
    attr.ClrObservable();
    BASE::InsEvent(index,attr);
  } 

  // InsUnobservableEvent(rName)
  TEMP Idx THIS::InsUnobservableEvent(const std::string& rName) {
    FD_DG("TcGenerator(" << this << ")::InsUnobservableEvent(" << rName << ")");
    EventAttr attr;
    attr.ClrObservable();
    return BASE::InsEvent(rName,attr);
  } 
    
  // SetObservable(index)
  TEMP void THIS::SetObservable(Idx index) {
    FD_DG("TcGenerator(" << this << ")::SetObservable(" << index << ")");
    EventAttr attr=BASE::EventAttribute(index);
    attr.SetObservable();
    BASE::mAlphabet.Attribute(index,attr);
  } 

  // SetObservable(rName)
  TEMP void THIS::SetObservable(const std::string& rName) {
    FD_DG("TcGenerator(" << this << ")::SetObservable(" << rName << ")");
    Idx index;
    try{
      index = BASE::EventIndex(rName);
    }
    catch (Exception) {
      std::stringstream errstr;
      errstr << "event \"" << rName << "\" not found in generator \"" << BASE::Name() << "\"";
      throw Exception("TcGenerator::SetObservable", errstr.str(), 202);
    }
    SetObservable(index);
  }

  //SetObservable(rEvents)
  TEMP void THIS::SetObservable(const EventSet& rEvents) {
    FD_DG("TcGenerator(" << this << ")::SetObservable(rEvents)");
    EventSet::Iterator it;
    for(it=rEvents.Begin(); it!=rEvents.End(); it++) {
      SetObservable(*it);
    }
  }
    
  // ClrObservable(index)
  TEMP void THIS::ClrObservable(Idx index) {
    FD_DG("TcGenerator(" << this << ")::ClrObservable(" << index << ")");
    EventAttr attr=BASE::EventAttribute(index);
    attr.ClrObservable();
    BASE::mAlphabet.Attribute(index,attr);
  } 

  // ClrObservable(rName)
  TEMP void THIS::ClrObservable(const std::string& rName) {
    FD_DG("TcGenerator(" << this << ")::ClrObservable(" << rName << ")");
    Idx index;
    try{
      index = BASE::EventIndex(rName);
    }
    catch (Exception) {
      std::stringstream errstr;
      errstr << "event \"" << rName << "\" not found in generator \"" << BASE::Name() << "\"";
      throw Exception("TcGenerator::ClrObservable", errstr.str(), 202);
    }
    ClrObservable(index);
  }

  //ClrObservable(rEvents)
  TEMP void THIS::ClrObservable(const EventSet& rEvents) {
    FD_DG("TcGenerator(" << this << ")::ClrObservable(rEvents)");
    EventSet::Iterator it;
    for(it=rEvents.Begin(); it!=rEvents.End(); it++) {
      ClrObservable(*it);
    }
  }

  //ObservableEvents()
  TEMP EventSet THIS::ObservableEvents(void) const {
    FD_DG("TcGenerator(" << this << ")::ObservableEvents()");
    EventSet res;
    EventSet::Iterator it;
    for(it=BASE::AlphabetBegin(); it!=BASE::AlphabetEnd(); it++) {
      if(Observable(*it)) res.Insert(*it);
    }
    return res;
  }
  
  //UnobservableEvents()
  TEMP
    EventSet THIS::UnobservableEvents(void) const {
    FD_DG("TcGenerator(" << this << ")::UnobservableEvents()");
    EventSet res;
    EventSet::Iterator it;
    for(it=BASE::AlphabetBegin(); it!=BASE::AlphabetEnd(); it++) {
      if(!Observable(*it)) res.Insert(*it);
    }
    return res;
  }
  

  // Forcible(index)
  TEMP bool THIS::Forcible(Idx index) const {
    EventAttr attr=BASE::EventAttribute(index);
    return attr.Forcible();
  } 
	
  // Forcible(rName)
  TEMP bool THIS::Forcible(const std::string& rName) const {
    EventAttr attr=BASE::EventAttribute(rName);
    return attr.Forcible();
  } 
	
  // InsForcibleEvent(index)
  TEMP void THIS::InsForcibleEvent(Idx index) {
    FD_DG("TcGenerator(" << this << ")::InsForcibleEvent(" << index << ")");
    EventAttr attr;
    attr.SetForcible();
    BASE::InsEvent(index,attr);
  } 

  // InsForcibleEvent(rName)
  TEMP Idx THIS::InsForcibleEvent(const std::string& rName) {
    FD_DG("TcGenerator(" << this << ")::InsForcibleEvent(" << rName << ")");
    EventAttr attr;
    attr.SetForcible();
    return BASE::InsEvent(rName,attr);
  } 

  // InsUnforcibleEvent(index)
  TEMP void THIS::InsUnforcibleEvent(Idx index) {
    FD_DG("TcGenerator(" << this << ")::InsUnforcibleEvent(" << index << ")");
    EventAttr attr;
    attr.ClrForcible();
    BASE::InsEvent(index,attr);
  } 

  // InsUnforcibleEvent(rName)
  TEMP Idx THIS::InsUnforcibleEvent(const std::string& rName) {
    FD_DG("TcGenerator(" << this << ")::InsUnforcibleEvent(" << rName << ")");
    EventAttr attr;
    attr.ClrForcible();
    return BASE::InsEvent(rName,attr);
  } 
    
  // SetForcible(index)
  TEMP void THIS::SetForcible(Idx index) {
    FD_DG("TcGenerator(" << this << ")::SetForcible(" << index << ")");
    EventAttr attr=BASE::EventAttribute(index);
    attr.SetForcible();
    BASE::mAlphabet.Attribute(index,attr);
  } 

  // SetForcible(rName)
  TEMP void THIS::SetForcible(const std::string& rName) {
    FD_DG("TcGenerator(" << this << ")::SetForcible(" << rName << ")");
    Idx index;
    try{
      index = BASE::EventIndex(rName);
    }
    catch (Exception) {
      std::stringstream errstr;
      errstr << "event \"" << rName << "\" not found in generator \"" << BASE::Name() << "\"";
      throw Exception("TcGenerator::SetForcible", errstr.str(), 202);
    }
    SetForcible(index);
  }

  //SetForcible(rEvents)
  TEMP void THIS::SetForcible(const EventSet& rEvents) {
    FD_DG("TcGenerator(" << this << ")::SetForcible(rEvents)");
    EventSet::Iterator it;
    for(it=rEvents.Begin(); it!=rEvents.End(); it++) {
      SetForcible(*it);
    }
  }
    
  // ClrForcible(index)
  TEMP void THIS::ClrForcible(Idx index) {
    FD_DG("TcGenerator(" << this << ")::ClrForcible(" << index << ")");
    EventAttr attr=BASE::EventAttribute(index);
    attr.ClrForcible();
    BASE::mAlphabet.Attribute(index,attr);
  } 

  // ClrForcible(rName)
  TEMP void THIS::ClrForcible(const std::string& rName) {
    FD_DG("TcGenerator(" << this << ")::ClrForcible(" << rName << ")");
    Idx index;
    try{
      index = BASE::EventIndex(rName);
    }
    catch (Exception) {
      std::stringstream errstr;
      errstr << "event \"" << rName << "\" not found in generator \"" << BASE::Name() << "\"";
      throw Exception("TcGenerator::ClrForcible", errstr.str(), 202);
    }
    ClrForcible(index);
  }

  //ClrForcible(rEvents)
  TEMP void THIS::ClrForcible(const EventSet& rEvents) {
    FD_DG("TcGenerator(" << this << ")::ClrForcible(rEvents)");
    EventSet::Iterator it;
    for(it=rEvents.Begin(); it!=rEvents.End(); it++) {
      ClrForcible(*it);
    }
  }

  //ForcibleEvents()
  TEMP EventSet THIS::ForcibleEvents(void) const {
    FD_DG("TcGenerator(" << this << ")::ForcibleEvents()");
    EventSet res;
    EventSet::Iterator it;
    for(it=BASE::AlphabetBegin(); it!=BASE::AlphabetEnd(); it++) {
      if(Forcible(*it)) res.Insert(*it);
    }
    return res;
  }
  
  //UnforcibleEvents()
  TEMP
    EventSet THIS::UnforcibleEvents(void) const {
    FD_DG("TcGenerator(" << this << ")::UnforcibleEvents()");
    EventSet res;
    EventSet::Iterator it;
    for(it=BASE::AlphabetBegin(); it!=BASE::AlphabetEnd(); it++) {
      if(!Forcible(*it)) res.Insert(*it);
    }
    return res;
  }
  
#undef TEMP
#undef BASE
#undef THIS

 
} // namespace faudes

#define CGENERATOR_H
#endif

