/** @file parallel.cpp parallel composition */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#include "parallel.h"

namespace faudes {

// Parallel(rGen1, rGen2, res)
void Parallel(const vGenerator& rGen1, const vGenerator& rGen2, vGenerator& rResGen) {
  // helpers:
  std::map< std::pair<Idx,Idx>, Idx> rcmap;
  // prepare result
  vGenerator* pResGen = &rResGen;
  if(&rResGen== &rGen1 || &rResGen== &rGen2) {
    pResGen= rResGen.NewP();
  }
  // doit
  Parallel(rGen1, rGen2, rcmap, *pResGen);
  if (rGen1.StateNamesEnabled() && rGen2.StateNamesEnabled() && rResGen.StateNamesEnabled()) 
    SetComposedStateNames(rGen1, rGen2, rcmap, *pResGen); 
  else
    pResGen->ClearStateNames();
  // copy result
  if(pResGen != &rResGen) {
    rResGen = *pResGen;
    delete pResGen;
  }
}
 

// Parallel(rGen1, rGen2, rReverseCompositionMap, res)
void Parallel(
  const vGenerator& rGen1, const vGenerator& rGen2, 
  std::map< std::pair<Idx,Idx>, Idx>& rReverseCompositionMap, 
  vGenerator& rResGen)
{
  FD_DF("Parallel(" << &rGen1 << "," << &rGen2 << ")");
  // prepare result
  vGenerator* pResGen = &rResGen;
  if(&rResGen== &rGen1 || &rResGen== &rGen2) {
    pResGen= rResGen.NewP();
  }
  pResGen->Clear();
  pResGen->Name(CollapsString(rGen1.Name()+"||"+rGen2.Name()));
  // rGen1.events() \ rGen2.events()
  EventSet sharedalphabet = rGen1.Alphabet() * rGen2.Alphabet();
  FD_DF("Parallel: shared events: " << sharedalphabet.ToString());

  // create res alphabet
  EventSet::Iterator eit;
  for (eit = rGen1.AlphabetBegin(); eit != rGen1.AlphabetEnd(); ++eit) {
    pResGen->InsEvent(*eit);
  }
  for (eit = rGen2.AlphabetBegin(); eit != rGen2.AlphabetEnd(); ++eit) {
    pResGen->InsEvent(*eit);
  }
  FD_DF("Parallel: inserted indices in rResGen.alphabet( "
      << pResGen->AlphabetToString() << ")");

  // todo stack
  std::stack< std::pair<Idx,Idx> > todo;
  // actual pair, new pair
  std::pair<Idx,Idx> currentstates, newstates;
  // state
  Idx tmpstate;  
  StateSet::Iterator lit1, lit2;
  TransSet::Iterator tit1, tit1_end, tit2, tit2_end;
  std::map< std::pair<Idx,Idx>, Idx>::iterator rcit;

  // push all combinations of initial states on todo stack
  FD_DF("Parallel: adding all combinations of initial states to todo:");
  for (lit1 = rGen1.InitStatesBegin(); lit1 != rGen1.InitStatesEnd(); ++lit1) {
    for (lit2 = rGen2.InitStatesBegin(); lit2 != rGen2.InitStatesEnd(); ++lit2) {
      currentstates = std::make_pair(*lit1, *lit2);
      todo.push(currentstates);
      rReverseCompositionMap[currentstates] = pResGen->InsInitState();
      FD_DF("Parallel:   (" << *lit1 << "|" << *lit2 << ") -> " 
          << rReverseCompositionMap[currentstates]);
    }
  }

  // start algorithm
  FD_DF("Parallel: processing reachable states:");
  while (! todo.empty()) {
    // get next reachable state from todo stack
    currentstates = todo.top();
    todo.pop();
    FD_DF("Parallel: processing (" << currentstates.first << "|" 
        << currentstates.second << ") -> " 
        << rReverseCompositionMap[currentstates]);
    // iterate over all rGen1 transitions 
    // (includes execution of shared events)
    tit1 = rGen1.TransRelBegin(currentstates.first);
    tit1_end = rGen1.TransRelEnd(currentstates.first);
    for (; tit1 != tit1_end; ++tit1) {
      // if event not shared
      if (! sharedalphabet.Exists(tit1->Ev)) {
        FD_DF("Parallel:   exists only in rGen1");
        newstates = std::make_pair(tit1->X2, currentstates.second);
        // add to todo list if composition state is new
        rcit = rReverseCompositionMap.find(newstates);
        if (rcit == rReverseCompositionMap.end()) {
          todo.push(newstates);
          tmpstate = pResGen->InsState();
          rReverseCompositionMap[newstates] = tmpstate;
          FD_DF("Parallel:   todo push: (" << newstates.first << "|" 
              << newstates.second << ") -> " 
              << rReverseCompositionMap[newstates]);
        }
        else {
          tmpstate = rcit->second;
        }
        pResGen->SetTransition(rReverseCompositionMap[currentstates], tit1->Ev, 
            tmpstate);
        FD_DF("Parallel:   add transition to new generator: " 
            << rReverseCompositionMap[currentstates] << "-" << tit1->Ev << "-" 
            << tmpstate);
      }
      // if shared event
      else {
        FD_DF("Parallel:   common event");
        // find shared transitions
        tit2 = rGen2.TransRelBegin(currentstates.second, tit1->Ev);
        tit2_end = rGen2.TransRelEnd(currentstates.second, tit1->Ev);
        for (; tit2 != tit2_end; ++tit2) {
          newstates = std::make_pair(tit1->X2, tit2->X2);
          // add to todo list if composition state is new
          rcit = rReverseCompositionMap.find(newstates);
          if (rcit == rReverseCompositionMap.end()) {
            todo.push(newstates);
            tmpstate = pResGen->InsState();
            rReverseCompositionMap[newstates] = tmpstate;
            FD_DF("Parallel:   todo push: (" << newstates.first << "|" 
                << newstates.second << ") -> " 
                << rReverseCompositionMap[newstates]);
          }
          else {
            tmpstate = rcit->second;
          }
          pResGen->SetTransition(rReverseCompositionMap[currentstates], 
              tit1->Ev, tmpstate);
          FD_DF("Parallel:   add transition to new generator: " 
              << rReverseCompositionMap[currentstates] << "-" 
              << tit1->Ev << "-" << tmpstate);
        }
      }
    }
    // iterate over all rGen2 transitions 
    // (without execution of shared events)
    tit2 = rGen2.TransRelBegin(currentstates.second);
    tit2_end = rGen2.TransRelEnd(currentstates.second);
    for (; tit2 != tit2_end; ++tit2) {
      if (! sharedalphabet.Exists(tit2->Ev)) {
        FD_DF("Parallel:   exists only in rGen2");
        newstates = std::make_pair(currentstates.first, tit2->X2);
        // add to todo list if composition state is new
        rcit = rReverseCompositionMap.find(newstates);
        if (rcit == rReverseCompositionMap.end()) {
          todo.push(newstates);
          tmpstate = pResGen->InsState();
          rReverseCompositionMap[newstates] = tmpstate;
          FD_DF("Parallel:   todo push: (" << newstates.first << "|" 
              << newstates.second << ") -> " 
              << rReverseCompositionMap[newstates]);
        }
        else {
          tmpstate = rcit->second;
        }
        pResGen->SetTransition(rReverseCompositionMap[currentstates], 
            tit2->Ev, tmpstate);
        FD_DF("Parallel:   add transition to new generator: " 
            << rReverseCompositionMap[currentstates] << "-" 
            << tit2->Ev << "-" << tmpstate);
      }
    }
  }

  // set marked states
  for (lit1 = rGen1.MarkedStatesBegin(); 
      lit1 != rGen1.MarkedStatesEnd(); ++lit1) {
    for (lit2 = rGen2.MarkedStatesBegin(); 
        lit2 != rGen2.MarkedStatesEnd(); ++lit2) {
      currentstates = std::make_pair(*lit1, *lit2);
      rcit = rReverseCompositionMap.find(currentstates);
      if (rcit != rReverseCompositionMap.end()) {
        pResGen->SetMarkedState(rcit->second);
      }
    }
  }
  FD_DF("Parallel: set marked states at the end: " 
      << pResGen->MarkedStatesToString());
  // copy result
  if(pResGen != &rResGen) {
    rResGen = *pResGen;
    delete pResGen;
  }
}


// SParallel(rGen1, rGen2, res)
void SParallel(const vGenerator& rGen1, const vGenerator& rGen2, vGenerator& rResGen) {
  std::map< std::pair<Idx,Idx>, Idx> rcmap;
  SParallel(rGen1, rGen2, rcmap, rResGen);
}


// SParallel(rGen1, rGen2, rReverseCompositionMap, res)
void SParallel(
  const vGenerator& rGen1, const vGenerator& rGen2, 
  std::map< std::pair<Idx,Idx>, Idx>& rReverseCompositionMap, 
  vGenerator& rResGen)
{
  FD_DF("SParallel(" << &rGen1 << "," << &rGen2 << ")");
  // prepare result
  vGenerator* pResGen = &rResGen;
  if(&rResGen== &rGen1 || &rResGen== &rGen2) {
    pResGen= rResGen.NewP();
  }
  pResGen->Clear();
  // shared alphabet
  pResGen->InjectAlphabet(rGen1.Alphabet() * rGen2.Alphabet());
  FD_DF("SParallel: shared alphabet: "
      << (rGen1.Alphabet() * rGen2.Alphabet()).ToString());

  // todo stack
  std::stack< std::pair<Idx,Idx> > todo;
  // actual pair, new pair
  std::pair<Idx,Idx> currentstates, newstates;
  // state
  Idx tmpstate;
  
  StateSet::Iterator lit1, lit2;
  TransSet::Iterator tit1, tit1_end, tit2, tit2_end;
  std::map< std::pair<Idx,Idx>, Idx>::iterator rcit;

  // push all combinations of initial states on todo stack
  FD_DF("SParallel: adding all combinations of initial states to todo:");
  for (lit1 = rGen1.InitStatesBegin(); 
      lit1 != rGen1.InitStatesEnd(); ++lit1) {
    for (lit2 = rGen2.InitStatesBegin(); 
        lit2 != rGen2.InitStatesEnd(); ++lit2) {
      currentstates = std::make_pair(*lit1, *lit2);
      todo.push(currentstates);
      rReverseCompositionMap[currentstates] = pResGen->InsInitState();
      FD_DF("SParallel:   (" << *lit1 << "|" << *lit2 << ") -> " 
          << rReverseCompositionMap[currentstates]);
    }
  }

  // start algorithm
  FD_DF("SParallel: processing reachable states:");
  while (! todo.empty()) {
    // get next reachable state from todo stack
    currentstates = todo.top();
    todo.pop();
    FD_DF("SParallel: processing (" << currentstates.first << "|" 
        << currentstates.second << ") -> " << rReverseCompositionMap[currentstates]);
    // iterate over all rGen1 transitions
    tit1 = rGen1.TransRelBegin(currentstates.first);
    tit1_end = rGen1.TransRelEnd(currentstates.first);
    tit2 = rGen2.TransRelBegin(currentstates.second);
    tit2_end = rGen2.TransRelEnd(currentstates.second);
    while ((tit1 != tit1_end) && (tit2 != tit2_end)) {
      // shared event
      if (tit1->Ev == tit2->Ev) {
        newstates = std::make_pair(tit1->X2, tit2->X2);
        // add to todo list if composition state is new
        rcit = rReverseCompositionMap.find(newstates);
        if (rcit == rReverseCompositionMap.end()) {
          todo.push(newstates);
          tmpstate = pResGen->InsState();
          rReverseCompositionMap[newstates] = tmpstate;
          FD_DF("SParallel: todo push: (" << newstates.first << "|" 
              << newstates.second << ") -> " << rReverseCompositionMap[newstates]);
        }
        else {
          tmpstate = rcit->second;
        }
        pResGen->SetTransition(rReverseCompositionMap[currentstates], tit1->Ev, tmpstate);
        FD_DF("SParallel: add transition to new generator: " 
            << rReverseCompositionMap[currentstates] << "-" << tit1->Ev << "-" << tmpstate);
        ++tit1;
        ++tit2;
      }
      // try resync tit1
      else if (tit1->Ev < tit2->Ev) {
        ++tit1;
      }
      // try resync tit2
      else if (tit1->Ev > tit2->Ev) {
        ++tit2;
      }
    }
  }

  // set marked states
  for (lit1 = rGen1.MarkedStatesBegin(); 
      lit1 != rGen1.MarkedStatesEnd(); ++lit1) {
    for (lit2 = rGen2.MarkedStatesBegin(); 
        lit2 != rGen2.MarkedStatesEnd(); ++lit2) {
      currentstates = std::make_pair(*lit1, *lit2);
      rcit = rReverseCompositionMap.find(currentstates);
      if (rcit != rReverseCompositionMap.end()) {
        pResGen->SetMarkedState(rcit->second);
      }
    }
  }
  FD_DF("SParallel: set marked states at the end: " 
      << pResGen->MarkedStatesToString());
  // copy result
  if(pResGen != &rResGen) {
    rResGen = *pResGen;
    delete pResGen;
  }
}




// SetParallelStateNames
void SetComposedStateNames(
  const vGenerator& rGen1, const vGenerator& rGen2, 
  std::map< std::pair<Idx,Idx>, Idx>& rReverseCompositionMap, 
  vGenerator& rGen12)
{
  std::map< std::pair<Idx,Idx>, Idx>::const_iterator rcit;
  for(rcit=rReverseCompositionMap.begin(); rcit!=rReverseCompositionMap.end(); rcit++) {
    Idx x1=rcit->first.first;
    Idx x2=rcit->first.second;
    Idx x12=rcit->second;
    if(!rGen12.ExistsState(x12)) continue;
    std::string name1= rGen1.StateName(x1);
    if(name1=="") name1=ToStringInteger(x1);
    std::string name2= rGen2.StateName(x2);
    if(name2=="") name1=ToStringInteger(x2);
    std::string name12= name1 + "|" + name2;
    name12=rGen12.UniqueStateName(name12);
    rGen12.StateName(x12,name12);
  }
}


} // name space
