/** @file parallel.h parallel composition */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#ifndef PARALLEL_H

#include "agenerator.h"
#include <stack>
#include <map>
#include <set>

namespace faudes {


/**
 * Parallel composition.
 *
 * Constructs the parallel composition of two generators, where shared events
 * are synchronised while non-shared events are executed independantly. 
 * The resulting generators alphabet is the union of the argument alphabets.
 * In this implementation, only accessible states are generated. 
 * On deterministic input this functions constructs a deterministic output.
 * See also Parallel(const vGenerator&,std::map< std::pair<Idx,Idx>, Idx>&,const vGenerator&, vGenerator&).
 *
 * @param rGen1
 *   First generator
 * @param rGen2
 *   Second generator
 * @param rResGen
 *   Reference to resulting parallel composition generator
 *
 * @ingroup GeneratorFunctions
 *
 */
void Parallel(const vGenerator& rGen1, const vGenerator& rGen2, vGenerator& rResGen);


/**
 * Parallel composition.
 *
 * See Parallel(const vGenerator&, const vGenerator&, vGenerator&).
 * This version fills given reverse composition map to map pairs of old states
 * to new states.
 *
 * @param rGen1
 *   First generator
 * @param rGen2
 *   Second generator
 * @param rReverseCompositionMap
 *   Reverse composition map (map< pair<Idx,Idx>, Idx>)
 * @param rResGen
 *   Reference to resulting parallel composition generator
 */
void Parallel(
    const vGenerator& rGen1, const vGenerator& rGen2,
    std::map< std::pair<Idx,Idx>, Idx>& rReverseCompositionMap, 
    vGenerator& rResGen);


/**
 * Shared event parallel composition.
 * 
 * Parallel composition that executes shared events only. The resulting
 * generators alphabet is the interscetion of the argument alphabets.
 * In this implementation, only accessible states are generated. 
 * Assumes deterministic input generators, result is deterministic.
 *
 * @param rGen1
 *   First generator
 * @param rGen2
 *   Second generator
 * @param rResGen
 *   Reference to resulting parallel composition generator
 *
 * @ingroup GeneratorFunctions
 */
void SParallel(const vGenerator& rGen1, const vGenerator& rGen2, vGenerator& rResGen);


/**
 * Shared event parallel composition.
 *
 * See SParallel(const vGenerator&, const vGenerator&, vGenerator&).
 * This version fills given reverse composition map to map pairs of old states
 * to new states.
 *
 * @param rGen1
 *   First generator
 * @param rGen2
 *   Second generator
 * @param rReverseCompositionMap
 *   Reverse composition map (map< pair<Idx,Idx>, Idx>)
 * @param rResGen
 *   Reference to resulting parallel composition generator
 */
void SParallel(
    const vGenerator& rGen1, const vGenerator& rGen2,
    std::map< std::pair<Idx,Idx>, Idx>& rReverseCompositionMap, 
    vGenerator& rResGen);

/**
 * Helper: uses reverse composition map to track state names
 * in a paralell composition. Purely cosmetic.
 *
 * @param rGen1
 *   First generator
 * @param rGen2
 *   Second generator
 * @param rReverseCompositionMap
 *   Reverse composition map (map< pair<Idx,Idx>, Idx>)
 * @param rGen12
 *   Reference to resulting parallel composition generator
 */
void SetComposedStateNames(
    const vGenerator& rGen1, const vGenerator& rGen2,
    std::map< std::pair<Idx,Idx>, Idx>& rReverseCompositionMap, 
    vGenerator& rGen12);



} // namespace faudes

#define PARALLEL_H
#endif 

