/** @file vgenerator.cpp Class vGenerator */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Copyright (C) 2007  Thomas Moor
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNES FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#include "vgenerator.h"


namespace faudes {

// msObjectCount (static) 
Idx vGenerator::msObjectCount = 0;

// constructor
vGenerator::vGenerator(void) : 
    // my name
    mMyName("Generator"),
    // std symboltables
    mpStateSymbolTable(&mStateSymbolTable),
    mpEventSymbolTable(GlobalEventSymbolTablep()),
    // other members
    mStateNamesEnabled(true)  
{
  FD_DG("vGenerator(" << this << ")::vGenerator()");
  // track generator objects
  msObjectCount++; 
  mId = msObjectCount;
}

// copy constructor
vGenerator::vGenerator(const vGenerator& rOtherGen) :
    // copy name
    mMyName(rOtherGen.mMyName),
    // copy symboltables
    mStateSymbolTable(rOtherGen.mStateSymbolTable),
    mpStateSymbolTable(&mStateSymbolTable),
    mpEventSymbolTable(rOtherGen.mpEventSymbolTable),
    // other members
    mStateNamesEnabled(rOtherGen.mStateNamesEnabled)  
{
  FD_DG("vGenerator(" << this << ")::vGenerator(" << &rOtherGen << ")");
  // track generator objects
  msObjectCount++; 
  mId = msObjectCount;
}


// Id()
Idx vGenerator::Id(void) const {
  return mId;
}

// Name(rName)
void vGenerator::Name(const std::string& rName) {
  FD_DV("vGenerator(" << this << ")::Name(\"" << rName << "\")");
  mMyName = rName;
}

// Name()
std::string vGenerator::Name(void) const {
  return mMyName;
}

// EventSymbolTablep() const
SymbolTable* vGenerator::EventSymbolTablep(void) const {
  return mpEventSymbolTable;
}

// GlobalEventSymbolTablep() const
SymbolTable* vGenerator::GlobalEventSymbolTablep(void) {
  return SymbolTable::GlobalEventSymbolTablep();
}

// EventSymbolTablep(pSymTab) 
void vGenerator::EventSymbolTablep(SymbolTable* pSymTab) {
  mpEventSymbolTable=pSymTab;
}

// EventSymbolTablep(rOtherGen) 
void vGenerator::EventSymbolTablep(const vGenerator& rOtherGen) {
  EventSymbolTablep(rOtherGen.EventSymbolTablep());
}

// EventIndex(rName)
Idx vGenerator::EventIndex(const std::string& rName) const {
  return mpEventSymbolTable->Index(rName);
}

// EventName(index)
std::string vGenerator::EventName(Idx index) const {
  return mpEventSymbolTable->Symbol(index);
}

// EventName(index, name)
void vGenerator::EventName(Idx index, const std::string& rName) {
  FD_DG("vGenerator(" << this << ")::EventName(" 
      << index << ",\"" << rName << "\")");
#ifdef FAUDES_CHECKED
  if (! ExistsEvent(index)) {
    std::stringstream errstr;
    errstr << "event \"" << index << "\" not found in generator \""
	   << Name() << "\"";
    throw Exception("vGenerator::EventName(name)", errstr.str(), 89);
  }
#endif
  mpEventSymbolTable->SetEntry(index, rName);
}

// UniqueEventName(rName)
std::string vGenerator::UniqueEventName(const std::string& rName) const {
  std::string name=rName;
  if(name=="") name="ev";
  return mpEventSymbolTable->UniqueSymbol(name) ;
}

// NewEventSet()
EventSet vGenerator::NewEventSet(void) const {
  EventSet res;
  res.SymbolTablep(mpEventSymbolTable);
  return res;
}

// NewEventSetp()
EventSet* vGenerator::NewEventSetp(void) const {
  EventSet* res = new EventSet();
  res->SymbolTablep(mpEventSymbolTable);
  return res;
}


// StateSymbolTable() const
const SymbolTable& vGenerator::StateSymbolTable(void) const {
  return mStateSymbolTable;
}

// StateSymbolTable(rSymTab) 
void vGenerator::StateSymbolTable(const SymbolTable& rSymTab) {
  mStateSymbolTable=rSymTab;
  mpStateSymbolTable=&mStateSymbolTable;
}

// StateIndex(rName)
Idx vGenerator::StateIndex(const std::string& rName) const {
   return mpStateSymbolTable->Index(rName);
}

// StateName(index)
std::string vGenerator::StateName(Idx index) const {
  return mpStateSymbolTable->Symbol(index);
}

// StateName(index, name)
void vGenerator::StateName(Idx index, const std::string& rName) {
  FD_DG("vGenerator(" << this << ")::StateName(" 
      << index << ",\"" << rName << "\")");
#ifdef FAUDES_CHECKED
  if (! ExistsState(index)) {
    std::stringstream errstr;
    errstr << "state name \"" << rName << "\" not found in generator \""
	   << Name() << "\"";
    throw Exception("vGenerator::StateName(name)", errstr.str(), 90);
  }
#endif
  mpStateSymbolTable->SetEntry(index, rName);
}


// ClearStateNames()
void vGenerator::ClearStateNames(void) {
  FD_DG("vGenerator(" << this << ")::ClearStateNames()");
  mpStateSymbolTable->Clear();
  mStateNamesEnabled = false;
}


// ClrStateName(index)
void vGenerator::ClrStateName(Idx index) {
  FD_DG("Generator(" << this << ")::ClrStateName(\"" << index << "\")");
#ifdef FAUDES_CHECKED
  if (! ExistsState(index)) {
    std::stringstream errstr;
    errstr << "state \"" << index << "\" not found in generator \""
	   << Name() << "\"";
    throw Exception("vGenerator::ClrStateName(name)", errstr.str(), 90);
  }
#endif
  mpStateSymbolTable->ClrEntry(index);
}

// ClrStateName(rName)
void vGenerator::ClrStateName(const std::string& rName) {
  FD_DG("vGenerator(" << this << ")::ClrStateName(\"" << rName << "\")");
  Idx index = StateIndex(rName);
  ClrStateName(index);
}


// StateNamesEnabled()
bool vGenerator::StateNamesEnabled(void) const {
  return mStateNamesEnabled;
}

// StateNamesEnabled(flag)
void vGenerator::StateNamesEnabled(bool flag) {
  mStateNamesEnabled = flag;
}

// SetDefaultStateNames()
void vGenerator::SetDefaultStateNames(void) {
  FD_DG("vGenerator(" << this << ")::SetDefaultStateNames()");
  StateSet::Iterator it;
  for (it = StatesBegin(); it != StatesEnd(); ++it) {
    mpStateSymbolTable->SetDefaultSymbol(*it);
  }
}

// EnforceStateNames(rTemplate)
void vGenerator::EnforceStateNames(const std::string& rTemplate) {
  FD_DG("vGenerator(" << this << ")::EnforceStateNames(temp)");
  StateSet::Iterator it;
  for (it = StatesBegin(); it != StatesEnd(); ++it) {
    if(StateName(*it)=="") {
      std::string name=UniqueStateName(rTemplate + "_1");
      StateName(*it,name);
    }
  }
}

// UniqueStateName(rName)
std::string vGenerator::UniqueStateName(const std::string& rName) const {
  std::string name=rName;
  if(name=="") name="st";
  return mpStateSymbolTable->UniqueSymbol(name) ;
}

// EStr(index)
std::string vGenerator::EStr(Idx index) const {
  return EventName(index);
}

// SStr(index)
std::string vGenerator::SStr(Idx index) const {
  std::string name = StateName(index);
  if(name == "") name=ToStringInteger(index);
  return name+"["+faudes::ToStringInteger(index)+"]";
}

// TStr(index)
std::string vGenerator::TStr(const Transition& trans) const {
  return SStr(trans.X1) + "--(" + EStr(trans.Ev) + ")-->" + SStr(trans.X2);
}



// GraphWrite(rFileName,rOutFormat)
void vGenerator::GraphWrite(const std::string& rFileName, const std::string& rOutFormat,
   const std::string& rDotExec) {
  FD_DG("vGenerator::GraohWrite(...)");
  // generate temp dot ibput file
  std::string dotin = CreateTempFile();
  if(dotin == "") {
    std::stringstream errstr;
    errstr << "Exception opening temp file";
    throw Exception("vGenerator::GraphWrite", errstr.str(), 2);
  }
  try{
    DotWrite(dotin);
  } 
  catch (faudes::Exception& exception) {  
    RemoveFile(dotin);
    std::stringstream errstr;
    errstr << "Exception writing dot input file";
    throw Exception("vGenerator::GraphWrite", errstr.str(), 2);
  }
  try{
    faudes::ProcessDot(dotin,rFileName,rOutFormat,rDotExec);
  } 
  catch (faudes::Exception& exception) {  
    RemoveFile(dotin);
    std::stringstream errstr;
    errstr << "Exception processing dot file";
    throw Exception("vGenerator::GraphWrite", errstr.str(), 3);
  }
  RemoveFile(dotin);
}


} // name space


